<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
	"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
    <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
        <head>
    <title>Iñtërnâtiônàlizætiøn</title>
</head>

<body>
    <h2>{{dude}} Iñtërnâtiônàlizætiøn</h2>
    <p>Τὸ πλοῖόν μου τὸ μετεωριζόμενον ἐστι<br />
πλῆρες ἐγχελέων.</p>
    </body>
</html>

